from measurement.measures import *
from measurement.utils import total_ordering


@total_ordering
class UnknownMeasure(object):
    def __init__(self, measure, original_unit, value):
        self.measure = measure
        self.original_unit = original_unit
        self.value = value

    def __getattr__(self, name):
        raise AttributeError(
            'UnknownMeasures cannot be convered to other units.'
        )

    def get_measurement_parts(self):
        return self.measure, self.original_unit, self.value

    def __eq__(self, other):
        if isinstance(other, UnknownMeasure):
            if self.measure == other.measure:
                if self.value == other.value:
                    return True
        return False

    def __lt__(self, other):
        if isinstance(other, UnknownMeasure):
            if self.measure == other.measure:
                if self.value < other.value:
                    return True
                return False
        return NotImplemented

    def __repr__(self):
        return '%s(?=%s)' % (self.measure, self.value)

    def __str__(self):
        return '%s ? (%s)' % (self.value, self.measure)



import re

from django import forms
from django.contrib.admin import ModelAdmin

from django_measurement.fields import MeasurementValueField
from django_measurement.utils import build_measure_list, guess_measurement


all_measures = build_measure_list()


class MeasurementAdminForm(forms.ModelForm):
    MEASURE_FIELD_NAME_RE = re.compile('^(.*)_value$')
    MEASURE_VALUE_SPLIT_RE = re.compile('^([0-9.-]+) ([a-zA-Z_/]+)$')
    MEASURE_FIELDS = {}

    def __init__(self, *args, **kwargs):
        instance = None
        if 'instance' in kwargs:
            instance = kwargs['instance']
        super(MeasurementAdminForm, self).__init__(*args, **kwargs)
        for field in self.Meta.model._meta.fields:
            if isinstance(field, MeasurementValueField):
                attr_name = self.MEASURE_FIELD_NAME_RE.match(
                    field.name
                ).group(1)
                self.MEASURE_FIELDS[attr_name] = {
                    'required': not field.blank
                }
        for field, fkwargs in self.MEASURE_FIELDS.items():
            choices = [(k, k) for k in all_measures.keys()]
            choices.insert(0, (None, ''))
            self.fields[field + '_measure'] = forms.ChoiceField(
                choices=choices,
                **fkwargs
            )
            self.fields[field] = forms.CharField(**fkwargs)
            if instance is not None:
                value = getattr(instance, field)
                self.initial[field + '_measure'] = value.__class__.__name__
                self.initial[field] = str(value)

    def clean(self):
        for field in self.MEASURE_FIELDS.keys():
            if field + '_measure' not in self.cleaned_data:
                continue
            measure_name = self.cleaned_data[field + '_measure']
            measure_value = self.cleaned_data.get(field, '')

            if not measure_value or not measure_name:
                continue

            measure = all_measures[measure_name]
            matcher = self.MEASURE_VALUE_SPLIT_RE.match(measure_value)
            try:
                unit = matcher.group(2).replace('/', '__')
                measurement = guess_measurement(
                    matcher.group(1), unit,
                    measures=[measure]
                )
            except (ValueError, AttributeError):
                raise forms.ValidationError(
                    '%s is not a valid measurement of %s' % (
                        measure_value,
                        measure_name.lower(),
                    )
                )
            self.cleaned_data.pop(field + '_measure')
            self.cleaned_data[field] = measurement
        return self.cleaned_data

    def save(self, *args, **kwargs):
        for field in self.MEASURE_FIELDS:
            setattr(
                self.instance,
                field,
                self.cleaned_data[field]
            )
        return super(MeasurementAdminForm, self).save(*args, **kwargs)


class MeasurementAdmin(ModelAdmin):
    form = MeasurementAdminForm

    def get_fieldsets(self, request, obj=None):
        form = self.get_form(request, obj)(instance=obj)
        return [(None, {'fields': form.fields.keys()})]



import logging
import re

from django.core.exceptions import ValidationError
from django.db.models import signals
from django.db.models.fields import CharField, Field, FloatField
from django import forms

from django_measurement import measure, utils


logger = logging.getLogger(__name__)


class OriginalUnitField(CharField):
    pass


class MeasureNameField(CharField):
    pass


class MeasurementValueField(FloatField):
    pass


class MeasurementFormField(forms.CharField):
    MEASUREMENT_RE = re.compile(r'(?P<value>[0-9.]+) ?(?P<unit>[A-Za-z_]+)')

    def to_python(self, value):
        if not value:
            return None
        match = self.MEASUREMENT_RE.match(value)
        if not match:
            raise ValidationError('%s could not be parsed into a known measurement' % value)
        value, unit = match.groups()
        unit = unit.lower()
        measurement = utils.guess_measurement(value, unit)
        if not match:
            raise ValidationError('%s is not a known unit' % unit)
        return measurement

def get_measurement_parts(value):
    if isinstance(value, measure.UnknownMeasure):
        return value.get_measurement_parts()
    measure_name = '%s(%s)' % (
        value.__class__.__name__,
        value.STANDARD_UNIT
    )
    original_unit = value.unit
    standard_value = value.standard
    return measure_name, original_unit, standard_value

class MeasurementFieldDescriptor(object):
    MEASUREMENT_NAME_RE = re.compile(r'([a-zA-Z0-9]+)(?:\(([a-zA-Z0-9_]+)\))?')

    def __init__(self, field, measurement_field_name, original_unit_field_name, measure_field_name):
        self.field = field
        self.measurement_field_name = measurement_field_name
        self.original_unit_field_name = original_unit_field_name
        self.measure_field_name = measure_field_name

    def _get_measure_by_name(self, measure_name):
        measures = utils.build_measure_list()
        return measures[measure_name]

    def _get_measure_name_and_std_unit(self, measure_string):
        return self.MEASUREMENT_NAME_RE.search(measure_string).groups()

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self

        measure_packed = getattr(
            instance, 
            self.measure_field_name
        )
        if measure_packed:
            measure_name, std_unit = self._get_measure_name_and_std_unit(
                    measure_packed
            )
        else:
            measure_name = ''
            std_unit = ''
        measurement_value = getattr(
            instance,
            self.measurement_field_name,
        )
        original_unit = getattr(
            instance,
            self.original_unit_field_name,
        )
        if not measure_name or not original_unit:
            return ''

        try:
            instance_measure = self._get_measure_by_name(measure_name)
            if std_unit and instance_measure.STANDARD_UNIT != std_unit:
                raise ValueError(
                    'Measurement %s base unit %s does not match stored %s' %
                    (
                        measure_name,
                        instance_measure.STANDARD_UNIT,
                        std_unit
                    )
                )
            measurement = utils.get_measurement(
                instance_measure,
                measurement_value,
                instance_measure.STANDARD_UNIT,
                original_unit=original_unit
            )
        except (AttributeError, KeyError, ImportError):
            measurement = measure.UnknownMeasure(
                measure=measure_name,
                original_unit=original_unit,
                value=measurement_value,
            )

        return measurement

    def __set__(self, instance, measurement):
        if instance is None:
            raise AttributeError("Must be accessed via instance")

        if measurement:
            measure, original_unit, standard_value = get_measurement_parts(
                measurement
            )
        else:
            measure = ''
            original_unit = ''
            standard_value = 0
        
        setattr(instance, self.measure_field_name, measure)
        setattr(instance, self.original_unit_field_name, original_unit)
        setattr(instance, self.measurement_field_name, standard_value)

class MeasurementField(Field):
    def __init__(self, 
            *args,
            **kwargs
        ):
        super(MeasurementField, self).__init__(*args, **kwargs)
        self.original_unit_field_name = '%s_unit'
        self.measure_field_name = '%s_measure'
        self.measurement_field_name = '%s_value'

    def get_original_unit_field_name(self):
        return self.original_unit_field_name % self.name

    def get_measure_field_name(self):
        return self.measure_field_name % self.name

    def get_measurement_field_name(self):
        return self.measurement_field_name % self.name

    def contribute_to_class(self, cls, name):
        self.name = name

        original_unit_field_name = self.get_original_unit_field_name()
        self.original_unit_field = OriginalUnitField(
            editable=False,
            blank=True,
            default='',
            max_length=50,
        )
        cls.add_to_class(original_unit_field_name, self.original_unit_field)

        measure_field_name = self.get_measure_field_name()
        self.measure_field = MeasureNameField(
            editable=False,
            blank=True,
            default='',
            max_length=255,
        )
        cls.add_to_class(measure_field_name, self.measure_field)

        measurement_field_name = self.get_measurement_field_name()
        self.measurement_field = MeasurementValueField(
            editable=False,
            blank=True,
            default='',
            max_length=50,
        )
        cls.add_to_class(measurement_field_name, self.measurement_field)

        field = MeasurementFieldDescriptor(
            self,
            measurement_field_name=measurement_field_name,
            original_unit_field_name=original_unit_field_name,
            measure_field_name=measure_field_name,
        )

        signals.pre_init.connect(self.instance_pre_init, sender=cls, weak=False)

        setattr(cls, self.name, field)

    def instance_pre_init(self, signal, sender, args, kwargs, **_kwargs):
        if self.name in kwargs:
            value = kwargs.pop(self.name)
            measure_name, original_unit, standard_value = get_measurement_parts(
                value
            )
            kwargs[self.get_measure_field_name()] = measure_name
            kwargs[self.get_original_unit_field_name()] = original_unit
            kwargs[self.get_measurement_field_name()] = standard_value

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ["^django_measurement\.fields\.MeasurementField"])
    add_introspection_rules([], ["^django_measurement\.fields\.OriginalUnitField"])
    add_introspection_rules([], ["^django_measurement\.fields\.MeasureNameField"])
    add_introspection_rules([], ["^django_measurement\.fields\.MeasurementValueField"])
except ImportError:
    pass



from django.conf import settings
from measurement.base import BidimensionalMeasure
from measurement.utils import get_all_measures, guess as guess_measurement

from django_measurement.measure import UnknownMeasure

MEASURE_OVERRIDES = getattr(settings, 'MEASURE_OVERRIDES', {})


def get_class_by_path(path):
    mod = __import__('.'.join(path.split('.')[:-1]))
    components = path.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def get_measure_unit_choices(include_measure=False):
    measures = build_measure_list()
    final_list = []
    for measure_name, measure in measures.items():
        if issubclass(measure, UnknownMeasure):
            continue
        if issubclass(measure, BidimensionalMeasure):
            continue
        measure_items = []
        for unit_name, _ in measure.UNITS.items():
            measure_items.append(
                ('%s.%s' % (measure_name, unit_name, ) if include_measure else unit_name, unit_name)
            )
        this_measure = tuple([measure_name, tuple(measure_items)])
        final_list.append(this_measure)
    return tuple(final_list)


def build_measure_list():
    all_measures = get_all_measures()
    measures = dict([(measure.__name__, measure) for measure in all_measures])
    for overridden_measure_name, cls_path in MEASURE_OVERRIDES.items():
        cls = get_class_by_path(
            cls_path
        )
        measures[overridden_measure_name] = cls
    # For unknown retrieved measures
    measures['UnknownMeasure'] = UnknownMeasure
    return measures


def get_measurement(measure, value, unit, original_unit=None):
    m = measure(
        **{unit: value}
    )
    if original_unit:
        m.unit = original_unit
    if isinstance(m, BidimensionalMeasure):
        m.reference.value = 1
    return m
